import flask
from typing import Union

class Request:
    def __init__(self, flask_request: flask.Request, app):
        self.flask_request = flask_request
        self.app = app
    
    def get_query_parameter(self, key: str) -> str:
        """Deprecated. Use params() instead."""
        print("WARNING: get_query_parameter() is deprecated. Use params() instead.")
        return self.params(key)

    def start_script(self, script_name: str, script_args: dict):
        return self.app.start_script(script_name, script_args)

    def store_dict(self, table_id: str, object_id: str, value: dict):
        return self.app.store_dict(table_id, object_id, value)
    
    def retrieve_dict(self, table_id: str, object_id: str) -> dict:
        return self.app.retrieve_dict(table_id, object_id)
    
    def delete_dict(self, table_id: str, object_id: str):
        return self.app.delete_dict(table_id, object_id)

    def query_dict(self, table_id: str, field_name: str, field_value) -> list:
        return self.app.query_dict(table_id, field_name, field_value)

    def list_object_ids(self, table_id: str) -> list:
        return self.app.list_object_ids(table_id)
    
    def list_objects(self, table_id: str) -> list:
        return self.app.list_objects(table_id)    
    
    def store_secret(self, secret_name: str, secret_value: str):
        return self.app.store_secret(secret_name, secret_value)
    
    def retrieve_secret(self, secret_name: str) -> str:
        return self.app.retrieve_secret(secret_name)

    def params(self, key: str = "") -> Union[str, dict]:
        """Returns the value of the query parameter with the given key. If no key is given, returns a dictionary of all query parameters."""
        if key == "":
            d = self.flask_request.args.to_dict()
            # print("d = ", d)

            if d:
                return {k: _sanitize(v) for k, v in d.items()}
            
            d = self.flask_request.form.to_dict()
            # print("d = ", d)

            if d:
                return {k: _sanitize(v) for k, v in d.items()}
            
            d = self.flask_request.values.to_dict()
            # print("d = ", d)

            if d:
                return {k: _sanitize(v) for k, v in d.items()}

        unsanitized = self.flask_request.args.get(key, '')

        if unsanitized is None or unsanitized == '':
            unsanitized = self.flask_request.form.get(key, '')

        if unsanitized is None or unsanitized == '':
            unsanitized = self.flask_request.values.get(key, '')

        # if unsanitized is None or unsanitized == '':
        #     unsanitized = self.flask_request.json.__getattribute__(key)

        return unsanitized.replace('<', '&lt;').replace('>', '&gt;') #.replace("'", "''").replace(";", "")

    def get_username(self) -> str:
        return flask.session.get('username')
    
def _sanitize(unsanitized: str) -> str:
    return unsanitized.replace('<', '&lt;').replace('>', '&gt;')



# Inspired by https://stackoverflow.com/a/40466535/3179416
from flask import Flask, Response
import flask
from .request import Request
from .all_components import *
from werkzeug.security import generate_password_hash, check_password_hash
import traceback

class LogoutHandler(object):
    def __init__(self, pycob_app):
        self.pycob_app = pycob_app
    
    def __call__(self, *args):
        flask.session.pop('username')
        return flask.redirect("/auth/login")

class LoginHandler(object):
    def __init__(self, pycob_app):
        self.pycob_app = pycob_app
    
    def __call__(self, *args):
        request = Request(flask.request, self.pycob_app)
        username = request.get_query_parameter("username")
        password = request.get_query_parameter("password")
        redirect = request.get_query_parameter("redirect")

        if redirect is None or redirect == "":
            redirect = "/auth/profile"

        user_data = self.pycob_app.retrieve_dict("users", username)

        if user_data is not None and 'password_hash' in user_data and check_password_hash(user_data['password_hash'], password):
            flask.session['username'] = username
            return flask.redirect(redirect)
        else:
            return flask.redirect("/auth/login_retry?redirect=" + redirect)

class SignupHandler(object):
    def __init__(self, pycob_app):
        self.pycob_app = pycob_app
    
    def __call__(self, *args):
        request = Request(flask.request, self.pycob_app)
        username = request.get_query_parameter("username")
        password = request.get_query_parameter("password")
        email = request.get_query_parameter("email")

        user_data = self.pycob_app.retrieve_dict("users", username)

        if user_data is not None:
            return flask.redirect("/auth/signup?message=" + "Username already exists. Please try again.")

        user_data = {
            "username": username,
            "password_hash": generate_password_hash(password),
            "email": email
        }

        self.pycob_app.store_dict("users", username, user_data)
        flask.session['username'] = username

        return flask.redirect("/auth/profile")

class PageHandler(object):
    def __init__(self, pycob_app, action, redirect_url, protect_with_code):
        self.action = action
        self.pycob_app = pycob_app
        self.redirect_url = redirect_url
        self.protect_with_code = protect_with_code
        self.response = Response(status=200, headers={})

    def __call__(self, *args):
        request = Request(flask.request, self.pycob_app)

        if self.pycob_app.error is not None:
            page = Page("Error")
            card = page.add_card()
            card.add_header("Error", size=5)
            card.add_text("An error occured while trying to load this page.")
            card.add_alert(str(self.pycob_app.error), "Error", color="red")
        else:            
            try:
                page = self.action(request)
            except Exception as e:
                page = Page("Error")
                card = page.add_card()
                card.add_header("Error", size=5)
                card.add_text("An error occured while trying to load this page.")
                card.add_alert(str(e), "Error", color="red")

                for i, x in enumerate(traceback.TracebackException.from_exception(e).format()):
                    if i > 1:
                        card.add_code(x, header="Traceback", prefix="")

        if page is None:
            raise ValueError(f'Did you forget to return the page at the end of the {self.action.__name__} function?')

        if self.redirect_url is not None:
            username = request.get_username()

            if username is None or username == "":                
                return flask.redirect("/auth/login?redirect=" + self.redirect_url)

        if self.protect_with_code is not None:
            login_code = request.get_query_parameter("login_code")
            
            if login_code is None or login_code != str(self.protect_with_code):
                page = Page("Enter Code")
                card = page.add_card()
                card.add_header("Enter Code", size=5)
                if login_code != "" and login_code != str(self.protect_with_code):
                    card.add_alert("Incorrect Code", "Error", color="red")
                card.add_text("Contact the developer of this site to give you a code for access")
                form = card.add_form(action="?")

                form.add_formtext("Code", "login_code", "Code from the developer")
                form.add_formsubmit('Enter')


        # if flask.request.accept_mimetypes['application/json'] and (not flask.request.accept_mimetypes['text/html']):
        #     json_response = page._to_json()
        #     print(json_response)
        #     return json_response, '200 OK', {'Content-Type': 'application/json'}

        html = ""

        html += _tailwind_header_to_sidebar(page.title, page.description, page.image, flask.request.url)

        if page.auto_navbar:
            html += get_navbar_html(self.pycob_app, request.get_username())

        # Add a sidebar here
        sidebar = _get_sidebar(page.components)
        if len(sidebar.components) > 0:
            html += '''<div class="flex">'''
            html += sidebar.to_html()

        html += page.to_html()

        if len(sidebar.components) > 0:
            html += "</div>"

        if page.auto_footer:
            html += get_footer_html(self.pycob_app)

        html += _tailwind_body_end

        return html

def get_navbar_html(pycob_app, username: str):
    if username == "" or username is None:
        navbar = NavbarComponent(pycob_app.name, "https://cdn.pycob.com/pycob_hex.png")
    else:
        navbar = NavbarComponent(pycob_app.name, "https://cdn.pycob.com/pycob_hex.png", button_label=username, button_url="/auth/profile")

    for page_path, page_dict in pycob_app.pages.items():
        if page_dict['show_in_navbar']:
            navbar.add_plainlink(page_dict['page_name'], "/" + page_path, "block rounded-lg py-2 pl-3 pr-4 text-white hover:bg-blue-800 md:p-2")

    return navbar.to_html()

def get_footer_html(pycob_app):
    footer = FooterComponent(pycob_app.name, pycob_app.subtitle, "https://cdn.pycob.com/pycob_hex.png")

    categorized = {}

    for page_path, page_dict in pycob_app.pages.items():
        if page_dict['footer_category'] is not None:
            if page_dict['footer_category'] in categorized:
                categorized[page_dict['footer_category']].append( FooterlinkComponent(page_dict['page_name'], "/" + page_path) )
            else:
                categorized[page_dict['footer_category']] = [ FooterlinkComponent(page_dict['page_name'], "/" + page_path) ]

    for category_name, footerlinks in categorized.items():
        footercategory = FootercategoryComponent(category_name, components=footerlinks)
        footer.add_component(footercategory)

    return footer.to_html()

def _get_sidebar(components) -> SidebarComponent:
    sidebar = SidebarComponent()

    sections = list(filter(lambda x: isinstance(x, SectionComponent), components))

    current_category = None

    for section in sections:
        if section.level == 1:
            if current_category is not None:
                sidebar.add_component(current_category)

            current_category = SidebarcategoryComponent(section.name)
        else:
            current_category.add_sidebarlink(section.name, "#" + section.id)
    
    if current_category is not None:
        sidebar.add_component(current_category)

    return sidebar

def _tailwind_header_to_sidebar(title: str, description: str, image: str, url: str) -> str:
    html = '''
    <!doctype html>
    <html>
    <head>
    <meta charset="UTF-8">
    <link rel="apple-touch-icon" sizes="180x180" href="https://cdn.pycob.com/apple-touch-icon.png">
    <link rel="icon" type="image/png" sizes="32x32" href="https://cdn.pycob.com/favicon-32x32.png">
    <link rel="icon" type="image/png" sizes="16x16" href="https://cdn.pycob.com/favicon-16x16.png">
    <link rel="mask-icon" href="https://cdn.pycob.com/safari-pinned-tab.svg" color="#5bbad5">
    <link rel="shortcut icon" href="https://cdn.pycob.com/favicon.ico">
    <meta name="msapplication-TileColor" content="#603cba">
    <meta name="msapplication-config" content="https://cdn.pycob.com/browserconfig.xml">
    <meta name="theme-color" content="#ffffff">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>''' + title + '''</title>
    <meta property="og:title" content="''' + title + '''">
    <meta property="og:description" content="''' + description + '''">
    <meta property="og:image" content="''' + image + '''">
    <meta property="og:url" content="''' + url + '''">
    <script src="https://cdn.tailwindcss.com"></script>
    <script src="https://unpkg.com/ag-grid-community/dist/ag-grid-community.min.js"></script>
    <script src="https://cdn.plot.ly/plotly-2.18.2.min.js"></script>
    <script>
        tailwind.config = {
            theme: {
                extend: {
                    colors: {
                        clifford: '#da373d',
                    }
                }
            },
            darkMode: 'class'
        }
    </script>
    <script>
        function toggleDarkMode() {
            if (document.documentElement.classList.contains('dark')) {
                document.documentElement.classList.remove('dark')
                document.getElementById("moon").classList.add("hidden")
                document.getElementById("sun").classList.remove("hidden")
            } else {
                document.documentElement.classList.add('dark')
                document.getElementById("sun").classList.add("hidden")
                document.getElementById("moon").classList.remove("hidden")
            }
        }

        function setLoading(el) {
            el.querySelectorAll("button[type=submit]").forEach((button) => {
                button.innerHTML = '<svg aria-hidden="true" role="status" class="inline w-4 h-4 mr-3 text-white animate-spin" viewBox="0 0 100 101" fill="none" xmlns="http://www.w3.org/2000/svg"><path d="M100 50.5908C100 78.2051 77.6142 100.591 50 100.591C22.3858 100.591 0 78.2051 0 50.5908C0 22.9766 22.3858 0.59082 50 0.59082C77.6142 0.59082 100 22.9766 100 50.5908ZM9.08144 50.5908C9.08144 73.1895 27.4013 91.5094 50 91.5094C72.5987 91.5094 90.9186 73.1895 90.9186 50.5908C90.9186 27.9921 72.5987 9.67226 50 9.67226C27.4013 9.67226 9.08144 27.9921 9.08144 50.5908Z" fill="#E5E7EB"/><path d="M93.9676 39.0409C96.393 38.4038 97.8624 35.9116 97.0079 33.5539C95.2932 28.8227 92.871 24.3692 89.8167 20.348C85.8452 15.1192 80.8826 10.7238 75.2124 7.41289C69.5422 4.10194 63.2754 1.94025 56.7698 1.05124C51.7666 0.367541 46.6976 0.446843 41.7345 1.27873C39.2613 1.69328 37.813 4.19778 38.4501 6.62326C39.0873 9.04874 41.5694 10.4717 44.0505 10.1071C47.8511 9.54855 51.7191 9.52689 55.5402 10.0491C60.8642 10.7766 65.9928 12.5457 70.6331 15.2552C75.2735 17.9648 79.3347 21.5619 82.5849 25.841C84.9175 28.9121 86.7997 32.2913 88.1811 35.8758C89.083 38.2158 91.5421 39.6781 93.9676 39.0409Z" fill="currentColor"/></svg>Loading...';
                button.disabled = true;
            });

            setTimeout(() => {
                el.querySelectorAll("button[type=submit]").forEach((button) => {
                    button.innerHTML = 'Try Again';
                    button.disabled = false;
                });
            }, "10000")
        }

        function toggleMore(button) {
            x = button

            if (x.innerHTML === "more") {
                x.innerHTML = "less";
                text_full = x.getAttribute("data-text-full")
                text_truncated = x.getAttribute("data-text-truncated")

                x.parentElement.innerHTML = text_full + ' <button data-text-full="' + text_full + '" data-text-truncated="' + text_truncated + '" onclick="toggleMore(this)" class="text-blue-500">less</button>'
            } else {
                x.innerHTML = "more";
                text_full = x.getAttribute("data-text-full")
                text_truncated = x.getAttribute("data-text-truncated")

                x.parentElement.innerHTML = text_truncated + ' <button data-text-full="' + text_full + '" data-text-truncated="' + text_truncated + '" onclick="toggleMore(this)" class="text-blue-500">more</button>'
            }
        }
    </script>
    <style>
    .gradient-background {
        background: linear-gradient(-45deg, #ee7752, #e73c7e, #23a6d5, #23d5ab);
        background-size: 400% 400%;
        animation: gradient 180s ease infinite;
    }

    .gradient-text {
        color: transparent;
        background: linear-gradient(-45deg, #ee7752, #e73c7e, #23a6d5, #23d5ab);
        background-clip: text;
        -webkit-background-clip: text;
        animation: gradient 2s ease infinite;
    }

    @keyframes gradient {
        0% {
            background-position: 0% 50%;
        }
        50% {
            background-position: 100% 50%;
        }
        100% {
            background-position: 0% 50%;
        }
    }

    </style>
    </head>
    <body class="flex flex-col h-screen dark:bg-gray-900 ">
'''
    return html



_tailwind_body_end = '''            
        </body>
        </html>
    '''


"""
# Services
## Store Dataframes
This works for any objects that can be pickled but the most common use case is for storing pandas dataframes.

`store_pickle`
`fetch_pickle`

## Store Dictionaries
`store_dict`
`fetch_dict`

## Secrets
`store_secret`
`fetch_secret`

## Background Jobs
You can also configure these to run on a schedule on https://www.pycob.com

`start_script`

## User Authentication
When using Pycob Cloud, there is a built in user authentication system. You can use this to store and retrieve user data.
An invisible 'token' is passed as a parameter to your app URL if a user is logged in. You can use this token to retrieve the user's data.

Your app will see URLs like this:

`https://my_app.pycob.app/page?my_param=hello&token=...`

`store_user_data`
`fetch_user_data`

## Other

`send_email`

# Deploy

`pycob.deploy`

"""
import os
import random
import string
import pickle
import requests
from pathlib import Path

from .app import *
from .request import *
from .component_interface import *
from .all_components import *

api_key = None
temp_dir = os.getcwd() + '/tmp/' + ''.join(random.choices(string.ascii_uppercase, k=5))

def send_email(from_email: str, to_email: str, subject: str, content: str) -> None:
    """
    ### Sends an email

    #### Example
    ```python
    send_email(
        from_email='user1@example.com', 
        to_email='user2@example.com', 
        subject='Hello', 
        content='Hello World!'
    )
    ```
    """
    global api_key
    if api_key is None:
        __set_api_key()

    email = {
        "from_email": from_email,
        "to_email": to_email,
        "subject": subject,
        "content": content,
    }
    __send_api_request("send_email", email, api_key)


def start_script(script_name: str, script_args: dict) -> str:
    """
    ### Starts a script with given script name and arguments

    :param script_name: Name of the script to be started
    :type script_name: str
    :param script_args: Arguments to be passed to the script
    :type script_args: dict
    :return: Returns the job ID of the started script
    :rtype: str

    #### Example
    ```python
    start_script(
        script_name='my_script.py',
        script_args={
            'arg1': 'value1',
            'arg2': 'value2'
        }
    )
    ```
    """
    global api_key
    if api_key is None:
        __set_api_key()

    script = {
        "script_name": script_name,
        "args": script_args,
    }
    rv = __send_api_request("run_job", script, api_key)

    if 'job_id' in rv:
        return rv['job_id']
    
    return ""

def store_pickle(data, filename: str):
    """
    ### Pickles a given data and stores it in the cloud with the given filename

    :param data: Data to be pickled
    :type data: any
    :param filename: Name of the file to be created and stored
    :type filename: str
    :return: None

    #### Example
    ```python
    cob.store_pickle(
        data=[1,2,3],
        filename='my_data.pkl'
    )
    ```
    ```python
    cob.store_pickle(
        data={'key': 'value'},
        filename='my_dict.pkl'
    )
    ```
    ```python
    cob.store_pickle(
        data=42,
        filename='my_integer.pkl'
    )
    ```
    """
    # Turn data into a pickle with the given filename

    Path(temp_dir).mkdir(parents=True, exist_ok=True)

    file = open(temp_dir + "/" + filename, 'wb')    
    pickle.dump(data, file)
    file.close()

    url = _url_for_file_storage(filename)

    if url == "":
        print("Error: Unable to store file")
        return
    
    # Put file with Content-Type: application/octet-stream
    file = open(temp_dir + "/" + filename, 'rb')
    requests.put(url, data=file, headers={'Content-Type': 'application/octet-stream'})
    file.close()

def fetch_pickle(filename: str):
    """
    ### Loads and returns pickled data from a file in cloud storage

    :param filename: Name of the pickled file to be loaded
    :type filename: str
    :return: Returns the loaded pickled data
    :rtype: Any

    #### Example
    ```python
    data = cob.fetch_pickle('my_data.pkl')
    ```
    ```python
    data = cob.fetch_pickle('path/to/my_data.pkl')
    ```
    ```python
    data = cob.fetch_pickle('data/my_data.pkl')
    ```
    """
    try:
        with open(temp_dir + "/" + filename, 'rb') as existing_file:
            data = pickle.load(existing_file)
            existing_file.close()
            return data
    except:
        pass

    # Get the pickle with the given filename

    url = _url_for_file_retrieval(filename)

    if url == "":
        print("Error: Unable to fetch file")
        return

    # Get file with Accept: application/octet-stream
    rv = requests.get(url, headers={'Accept': 'application/octet-stream'})
    Path(temp_dir).mkdir(parents=True, exist_ok=True)
    file = open(temp_dir + "/" + filename, 'wb')
    file.write(rv.content)
    file.close()

    file = open(temp_dir + "/" + filename, 'rb')
    data = pickle.load(file)
    file.close()

    return data


def _url_for_file_storage(file_name: str) -> str:
    global api_key
    if api_key is None:
        __set_api_key()

    file = {
        "filename": file_name,
    }
    rv = __send_api_request("store_file", file, api_key)

    if 'url' in rv:
        return rv['url']

    return ""

def _url_for_file_retrieval(file_name: str) -> str:
    global api_key
    if api_key is None:
        __set_api_key()

    file = {
        "filename": file_name,
    }
    rv = __send_api_request("retrieve_file", file, api_key)

    if 'url' in rv:
        return rv['url']

    return ""

def store_dict(table_id: str, object_id: str, value: dict):
    """
    ### Stores a dictionary in the database

    :param table_id: Identifier of the table in which the object has to be stored
    :type table_id: str
    :param object_id: Identifier of the object in the table
    :type object_id: str
    :param value: Dictionary to be stored
    :type value: dict
    :return: None

    #### Example
    ```python
    store_dict(
        table_id='my_table',
        object_id='my_object',
        value={
            'key1': 'value1',
            'key2': 2,
            'key3': [1, 2, 3]
        }
    )
    ```
    """
    global api_key
    if api_key is None:
        __set_api_key()

    type_name = type(value).__name__
    print("Attempting to save a value of type = ", type_name)
    is_int = "int" in type_name
    is_float = "float" in type_name
    is_str = "str" in type_name
    is_list = "list" in type_name


    to_store = {"table_id": table_id, "object_id": object_id}

    if is_int or is_float or is_str:
        print("Saving flat value to DB")
        to_store['flat_value'] = value
    elif is_list:
        print("Saving list to DB")
        to_store['object_array'] = value
    else:
        print("Saving object to DB")
        to_store['object'] = value

    __send_api_request("store_object", to_store, api_key)

def fetch_dict(table_id: str, object_id: str) -> dict:
    """
    ### fetch a dictionary
    
    :param table_id: ID of the table to fetch the object from
    :type table_id: str
    :param object_id: ID of the object to be fetchd
    :type object_id: str
    :return: Returns the object as a dictionary
    :rtype: dict
    
    #### Example
    ```python
    data = cob.fetch_dict(
        table_id='table_123',
        object_id='object_456'
    )
    ```
    """
    global api_key
    if api_key is None:
        __set_api_key()

    server_response = __send_api_request("fetch_object", {"object_id": object_id, "table_id": table_id}, api_key)

    if 'object' in server_response:
        return server_response['object']

def delete_dict(table_id: str, object_id: str):
    """
    ### Deletes the object with given object ID from the table with given table ID

    :param table_id: ID of the table
    :type table_id: str
    :param object_id: ID of the object to be deleted
    :type object_id: str
    :return: None
    :rtype: None

    #### Example
    ```python
    cob.delete_dict(table_id='table_1', object_id='obj_1')
    ```
    """
    global api_key
    if api_key is None:
        __set_api_key()

    __send_api_request("delete_object", {"object_id": object_id, "table_id": table_id}, api_key)        

def query_dict(table_id: str, field_name: str, field_value) -> list:
    """
    ### Queries a table in the database and returns a list of objects matching the given field-value pair

    :param table_id: ID of the table to be queried
    :type table_id: str
    :param field_name: Name of the field to be queried
    :type field_name: str
    :param field_value: Value of the field to be queried
    :param: field_value: str or int
    :return: Returns a list of objects matching the given field-value pair
    :rtype: list of dicts

    #### Example
    ```python
    cob.query_dict(
        table_id='my_table',
        field_name='name',
        field_value='John'
    )
    ```
    """    
    global api_key
    if api_key is None:
        __set_api_key()

    server_response = __send_api_request("query_objects", {"field_name": field_name, "table_id": table_id, "field_value": field_value}, api_key)

    if type(server_response) is list:
        arr = list(map(lambda x:  x['object'] if 'object' in x else None, server_response))
        num_items = len(arr)
        print(f"Got {num_items} items from the database table {table_id} where {field_name} == {field_value}")
        return arr

def list_objects(table_id: str) -> list:
    """
    ### Get a list of objects from a table

    This function fetchs a list of objects from a table using the provided table ID.

    :param table_id: ID of the table from which to fetch the objects
    :type table_id: str
    :return: Returns a list of objects from the specified table
    :rtype: list

    #### Example
    ```python
    cob.list_objects(table_id='my_table')
    ```
    """
    global api_key
    if api_key is None:
        __set_api_key()

    server_response = __send_api_request("list_objects", {"table_id": table_id}, api_key)

    if type(server_response) is list:
        arr = list(map(lambda x:  x['object'] if 'object' in x else None, server_response))
        num_items = len(arr)
        print(f"Got {num_items} total items from the database table {table_id}")
        return arr

def list_object_ids(table_id: str) -> list:
    """
    ### Lists object IDs of a table

    :param table_id: ID of the table
    :type table_id: str
    :return: Returns a list of object IDs of the given table
    :rtype: list

    #### Example
    ```python
    cob.list_object_ids(table_id='my_table')
    ```
    """
    global api_key
    if api_key is None:
        __set_api_key()

    server_response = __send_api_request("list_object_ids", {"table_id": table_id}, api_key)

    if server_response is not None and 'object_ids' in server_response:
        return server_response['object_ids']

def store_secret(secret_name: str, secret_value: str):
    """
    ### Stores a secret with given name and value

    :param secret_name: Name of the secret to be stored
    :type secret_name: str
    :param secret_value: Value of the secret to be stored
    :type secret_value: str
    :return: None
    :rtype: None

    #### Example
    ```python
    cob.store_secret(
        secret_name='my_secret',
        secret_value='my_secret_value'
    )
    ```
    """
    store_dict("_secrets", secret_name, {"value": secret_value})

def fetch_secret(secret_name: str) -> str:
    """
    ### fetchs the value of a secret 

    :param secret_name: Name of the secret to be fetchd
    :type secret_name: str
    :return: Returns the value of the secret if it exists, otherwise None
    :rtype: str

    #### Example
    ```python
    cob.fetch_secret("my_secret")
    ``` 
    """
    try:
        return fetch_dict("_secrets", secret_name)['value']
    except:
        return None

def get_quota_status() -> dict:
    global api_key
    if api_key is None:
        __set_api_key()

    server_response = __send_api_request("get_quota_status", {}, api_key)

    return server_response

def __set_api_key():
    global api_key
    if api_key is None:
        # Check PYCOB_API_KEY environment variable
        api_key = os.environ.get("PYCOB_API_KEY")

        if api_key is None:

            if Path("pycob_api_key.txt").is_file():
                # Check for api_key.txt file
                f = open("pycob_api_key.txt", "r")
                api_key = f.read()

            if api_key is None:
                print("No API key found. ")
                print("Go to https://www.pycob.com to get an API key.")
                # Exit
                error = 'No API key found. Go to <a href="https://www.pycob.com">https://www.pycob.com</a> to get an API key.'

def __send_api_request(endpoint: str, data: dict, api_key: str) -> dict:
    response = requests.post("https://api.pycob.com/rpc/"+endpoint, json=data, headers={"PYCOB-API-KEY": api_key})
    
    if response.status_code != 200:
        error_str = "API request failed with status code " + str(response.status_code)
        print(error_str)
        return {"error": error_str}

    try:
        json_response = response.json()
        # print("API request returned: " + str(json_response))
        # print("Class of json_response: " + str(type(json_response)))
        return json_response
    except:
        print("API request failed to return valid JSON.")
        return {"error": "API request failed to return valid JSON."}